import pytest
import tempfile

from runner import main as runner
from differ import main as differ
from test_utils import get_expected_path

TOOL = 'ttxn'

OTF_FONT = 'SourceSansPro-Black_subset.otf'
TTF_FONT = 'SourceSansPro-Black_subset.ttf'
OTF2_FONT = 'SourceSansPro-ExtraLightIt.otf'
OTF3_FONT = 'SourceSansPro-Light.otf'
TTF2_FONT = 'NotoNaskhArabic-Regular.ttf'
TTF3_FONT = 'NotoNastaliqUrdu-Regular.ttf'


# -----
# Tests
# -----

@pytest.mark.parametrize('args, font_filename, exp_filename', [
    ([], OTF_FONT, 'OTF.ttx'),
    ([], TTF_FONT, 'TTF.ttx'),
    (['nh'], OTF_FONT, 'OTF_no_hints.ttx'),
    (['tGPOS'], OTF_FONT, 'OTF_GPOS_only.ttx'),
    (['tGSUB'], OTF_FONT, 'OTF_GSUB_only.ttx'),
    (['tGPOS'], OTF2_FONT, 'OTF2_GPOS_only.ttx'),
    (['tGSUB'], OTF2_FONT, 'OTF2_GSUB_only.ttx'),
    (['tGPOS'], OTF3_FONT, 'OTF3_GPOS_only.ttx'),
    (['tGSUB'], OTF3_FONT, 'OTF3_GSUB_only.ttx'),
    (['tGPOS'], TTF2_FONT, 'TTF2_GPOS_only.ttx'),
    (['tGSUB'], TTF2_FONT, 'TTF2_GSUB_only.ttx'),
    (['tGPOS', 'se'], TTF3_FONT, 'TTF3_GPOS_only.ttx'),
    (['tGSUB', 'se'], TTF3_FONT, 'TTF3_GSUB_only.ttx'),
])
def test_dump(args, font_filename, exp_filename):
    save_path = tempfile.mkstemp()[1]
    runner(['-t', TOOL, '-f', font_filename, '-o',
            f'o{save_path}'] + args)
    expected_path = get_expected_path(exp_filename)
    assert differ([expected_path, save_path, '-s', '<ttFont sfntVersion'])



# pdfdoc.py
"""
PDFgen is a library to generate PDF files containing text and graphics.  It is the
foundation for a complete reporting solution in Python.

The module pdfdoc.py handles the 'outer structure' of PDF documents, ensuring that
all objects are properly cross-referenced and indexed to the nearest byte.  The
'inner structure' - the page descriptions - are presumed to be generated before
each page is saved.
pdfgen.py calls this and provides a 'canvas' object to handle page marking operators.
piddlePDF calls pdfgen and offers a high-level interface.

(C) Copyright Andy Robinson 1998-1999

Modified 7/25/2006 read rooberts. Added supported for embedding fonts.
"""

import sys
import time
import zlib

from afdko.pdflib import pdfutils
from afdko.pdflib.pdfutils import LINEEND   # this constant needed in both

Log = sys.stderr  # reassign this if you don't want err output to console

##############################################################
#
#            Constants and declarations
#
##############################################################


StandardEnglishFonts = [
    'Courier', 'Courier-Bold', 'Courier-Oblique', 'Courier-BoldOblique',
    'Helvetica', 'Helvetica-Bold', 'Helvetica-Oblique',
    'Helvetica-BoldOblique',
    'Times-Roman', 'Times-Bold', 'Times-Italic', 'Times-BoldItalic',
    'Symbol','ZapfDingbats']

kDefaultEncoding = '/MacRoman'
kDefaultFontType = "/Type1"
AFMDIR = '.'

A4 = (595.27,841.89)   #default page size


class PDFError(KeyError):
    pass


class PDFDocument:
    """Responsible for linking and writing out the whole document.
    Builds up a list of objects using add(key, object).  Each of these
    must inherit from PDFObject and be able to write itself into the file.
    For cross-linking, it provides getPosition(key) which tells you where
    another object is, or raises a KeyError if not found.  The rule is that
    objects should only refer ones previously written to file.
    """
    def __init__(self):
        self.objects = []
        self.objectPositions = {}

        self.pages = []
        self.pagepositions = []

        # position 1
        cat = PDFCatalog()
        cat.RefPages = 3
        cat.RefOutlines = 2
        self.add('Catalog', cat)

        # position 2 - outlines
        outl = PDFOutline()
        self.add('Outline', outl)

        # position 3 - pages collection
        self.PageCol = PDFPageCollection()
        self.add('PagesTreeRoot',self.PageCol)

        # mapping of Postscript font names to internal ones;
        # add all the standard built-in fonts.
        self.fonts = StandardEnglishFonts # list of PS names. MakeType1Fonts()
        self.fontDescriptors = {} # Initally, we have no font descriptors, as these are required only
                                  # for fonts other than the built-in fonts.
        self.fontMapping = {}
        fontIndex = 0
        for i in range(len(self.fonts)):
            psName = self.fonts[i]
            fontIndex = i+1
            encoding = kDefaultEncoding
            clientCtx = None
            getFontDescriptor = None
            pdfFont = self.MakeType1Font(fontIndex, psName, encoding, clientCtx, getFontDescriptor)
            self.add('Font.'+pdfFont.keyname, pdfFont)
            objectNumber = len(self.objects)
            self.fontMapping[psName+repr(encoding)] = [fontIndex, pdfFont, objectNumber, psName, encoding]
        self.fontdict = MakeFontDictionary(self.fontMapping) # This needs to be called again whenever a font is added.


        # position 17 - Info
        self.info = PDFInfo()  #hang onto it!
        self.add('Info', self.info)
        self.infopos = len(self.objects)  #1-based, this gives its position


    def add(self, key, obj):
        self.objectPositions[key] = len(self.objects)  # its position
        self.objects.append(obj)
        obj.doc = self
        return len(self.objects) - 1  # give its position

    def getPosition(self, key):
        """Tell you where the given object is in the file - used for
        cross-linking; an object can call self.doc.getPosition("Page001")
        to find out where the object keyed under "Page001" is stored."""
        return self.objectPositions[key]

    def setTitle(self, title):
        "embeds in PDF file"
        self.info.title = title

    def setAuthor(self, author):
        "embedded in PDF file"
        self.info.author = author

    def setSubject(self, subject):
        "embeds in PDF file"
        self.info.subject = subject

    def printXref(self):
        self.startxref = sys.stdout.tell()
        Log.write('xref\n')
        Log.write("%s %s" % (0, len(self.objects) + 1))
        Log.write('0000000000 65535 f')
        for pos in self.xref:
            Log.write('%0.10d 00000 n\n' % pos)

    def writeXref(self, f):
        self.startxref = f.tell()
        f.write(('xref' + LINEEND).encode('utf-8'))
        f.write(('0 %d' % (len(self.objects) + 1) + LINEEND).encode('utf-8'))
        f.write(('0000000000 65535 f' + LINEEND).encode('utf-8'))
        for pos in self.xref:
            f.write(('%0.10d 00000 n' % pos + LINEEND).encode('utf-8'))

    def printTrailer(self):
        print('trailer')
        print('<< /Size %d /Root %d 0 R /Info %d 0 R>>' % (
            len(self.objects) + 1, 1, self.infopos))
        print('startxref')
        print(self.startxref)

    def writeTrailer(self, f):
        f.write(('trailer' + LINEEND).encode('utf-8'))
        f.write(('<< /Size %d /Root %d 0 R /Info %d 0 R>>' % (len(self.objects) + 1, 1, self.infopos)  + LINEEND).encode('utf-8'))
        f.write(('startxref' + LINEEND).encode('utf-8'))
        f.write((str(self.startxref)  + LINEEND).encode('utf-8'))

    def SaveToFile(self, filename):
        with open(filename, 'wb') as fileobj:
            self.SaveToFileObject(fileobj)

    def SaveToFileObject(self, fileobj):
        """Open a file, and ask each object in turn to write itself to
        the file.  Keep track of the file position at each point for
        use in the index at the end"""
        f = fileobj
        self.xref = []
        f.write(("%PDF-1.2" + LINEEND).encode('utf-8'))  # for CID support
        f.write(b"%\xed\xec\xb6\xbe\r\n")
        for i, obj in enumerate(self.objects, 1):
            pos = f.tell()
            self.xref.append(pos)
            f.write((str(i) + ' 0 obj' + LINEEND).encode('utf-8'))
            obj.save(f)
            f.write(('endobj' + LINEEND).encode('utf-8'))
        self.writeXref(f)
        self.writeTrailer(f)
        f.write(('%%EOF').encode('utf-8'))  # no lineend needed on this one!

    def printPDF(self):
        "prints it to standard output.  Logs positions for doing trailer"
        print("%PDF-1.0")
        print("%\xed\xec\xb6\xbe")
        i = 1
        self.xref = []
        for obj in self.objects:
            pos = sys.stdout.tell()
            self.xref.append(pos)
            print(i, '0 obj')
            obj.printPDF()
            print('endobj')
            i = i + 1
        self.printXref()
        self.printTrailer()
        print("%%EOF",)

    def addPage(self, page):
        """adds page and stream at end.  Maintains pages list"""
        #page.buildstream()
        pos = len(self.objects) # work out where added

        page.ParentPos = 3   #pages collection
        page.info = {
            'parentpos':3,
            'fontdict':self.fontdict,
            'contentspos':pos + 2,
            }

        self.PageCol.PageList.append(pos+1)
        self.add('Page%06d'% len(self.PageCol.PageList), page)
        #self.objects.append(page)
        self.add('PageStream%06d'% len(self.PageCol.PageList), page.stream)
        #self.objects.append(page.stream)

    def hasFont(self, psfontname, encoding=kDefaultEncoding):
        if encoding is None:
            encoding = kDefaultEncoding
        try:
           entry = self.fontMapping[psfontname+repr(encoding)]
           if entry[4] == encoding:
              return 1
           else:
              return 0
        except KeyError:
           return 0

    def addFont(self, psname, encoding=kDefaultEncoding, clientCtx=None, getFontDescriptorItems=None, getEncodingInfo=None):
        """ If you pass in just the psname, you can font dict that will work
        with only one of the built-in fonts. To emebed a font, you need to pass
        in all of the rest. "encoding" must be a string that is either one of
        the stanrd PDF encoding names, like '/MacRoman', or a string containing
        a well-formed PDF encoding dict.

        clientCtx is a client-provided Python object that cntains the data used
        be the client call back functions, getFontDescriptorItems,
        getEncodingInfo.

        getFontDescriptorItems is a client call back function that must return
        fdText, type, fontStream, fontStreamType

        fdText must be sttring containing a well-formed /FontDescriptor dict. In
        this string, the the reference to the embedded font stream must be as an
        indirect reference, with the object number being a string format , ie.
        /FontFile3 %s 0 R" This is is necessary ot allow the font stream object
        number ot be filled in later, by the line: fdText = fdText %
        objectNumber in MakeType1Font.

        type must tbe the PDF name for the font type, such as "/Type1"

        fontStream must be the actual font data to be embedded, not the entire
        PDF stream object. At the moment, it supports only CFF font data.

        fontStreamType is the value for the stream font SubType, e.g. /Type1C
        getEncodingInfo is a client call back function that must return
        firstChar, lastChar, widths

        firstChar is the index in the encoding array of the first glyph name
        which is not notdef lastChar is the index in the encoding array of the
        last glyph name whcih is not notdef widths is the width array of the
        encoded glyphs. It must contain a list of widths for the glyphs in the
        encoding array from firstChar to lastChar.

        The call to addFont will add the font descriptor object and the embedded
        stream object the first time it is called. If it is called again with
        the same PS name and encoding string, it will simply reference the font
        descriptor object and the embedded stream object that was already added,
        but will create a new font dict with a new widths and encoding dict.


        """
        if self.hasFont(psname, encoding):
           return
        numFonts = len(self.fontMapping)
        fontIndex = numFonts+1
        pdfFont = self.MakeType1Font(fontIndex, psname, encoding, clientCtx, getFontDescriptorItems, getEncodingInfo)
        self.add('Font.'+pdfFont.keyname, pdfFont)
        objIndex = len(self.objects)
        self.fontMapping[psname+repr(encoding)] = [fontIndex, pdfFont, objIndex, psname, encoding]
        self.fontdict = MakeFontDictionary(self.fontMapping)

    def getInternalFontName(self, psfontname, encoding=kDefaultEncoding):
        if encoding is None:
            encoding = kDefaultEncoding
        try:
            entry = self.fontMapping[psfontname+repr(encoding)]
            pdfFont = entry[1]
            return "/%s" % (pdfFont.keyname)
        except:
            raise PDFError("Font %s not available in document" % psfontname)

    def getAvailableFonts(self):
        # There may be more
        entries = self.fontMapping.values()
        fontEntries = map(lambda entry: [entry[3], entry[4]], entries) # [psName, encoding]
        fontEntries.sort()
        return fontEntries

    def MakeType1Font(self, fontIndex, psName, encoding=kDefaultEncoding, clientCtx=None, getFontDescriptorItems=None, getEncodingInfo=None):
        "returns a font object"
        objectNumber = None
        type = kDefaultFontType
        firstChar = lastChar = widths = None
         # If there is no getFontDescriptoItems function, we will assume that this font is a built-in font,
         # and hence does not need a font descriptor.
        if getFontDescriptorItems:
            if getEncodingInfo:
                firstChar, lastChar, widths = getEncodingInfo(clientCtx)
            try:
                # check if we have already seen this
                objectNumber,type = self.fontDescriptors[psName]
            except KeyError:
                # Need to create a font descriptor object.
                fdText, type, formatName, fontStream, fontStreamType = getFontDescriptorItems(clientCtx)
                fontStreamObj = PDFEmbddedFont(fontStreamType)
                fontStreamObj.setStream(fontStream)
                self.add(formatName, fontStreamObj)
                objectNumber = len(self.objects)
                fdText = fdText % objectNumber
                fontDescriptorObj = PDFontDescriptor(fdText)
                self.add('FontDescriptor', fontDescriptorObj)
                objectNumber = len(self.objects)
                self.fontDescriptors[psName] = (objectNumber, type)
        if type == "/Type0":
            font = PDFType0Font('F'+str(fontIndex), psName, encoding, objectNumber, type, firstChar, lastChar, widths)
        else:
            font = PDFType1Font('F'+str(fontIndex), psName, encoding, objectNumber, type, firstChar, lastChar, widths)
        return font


##############################################################
#
#            Utilities
#
##############################################################

class OutputGrabber:
    """At times we need to put something in the place of standard
    output.  This grabs stdout, keeps the data, and releases stdout
    when done.

    NOT working well enough!"""
    def __init__(self):
        self.oldoutput = sys.stdout
        sys.stdout = self
        self.closed = 0
        self.data = []

    def write(self, x):
        if not self.closed:
            self.data.append(x)

    def getData(self):
        return ''.join(self.data)

    def close(self):
        sys.stdout = self.oldoutput
        self.closed = 1

    def __del__(self):
        if not self.closed:
            self.close()


def testOutputGrabber():
    gr = OutputGrabber()
    for i in range(10):
        print('line', i)
    data = gr.getData()
    gr.close()
    print('Data...', data)


##############################################################
#
#            PDF Object Hierarchy
#
##############################################################



class PDFObject:
    """Base class for all PDF objects.  In PDF, precise measurement
    of file offsets is essential, so the usual trick of just printing
    and redirecting output has proved to give different behaviour on
    Mac and Windows.  While it might be soluble, I'm taking charge
    of line ends at the binary level and explicitly writing to a file.
    The LINEEND constant lets me try CR, LF and CRLF easily to help
    pin down the problem."""
    def save(self, file):
        "Save its content to an open file"
        file.write('% base PDF object' + LINEEND)
    def printPDF(self):
        self.save(sys.stdout)


class PDFLiteral(PDFObject):
    " a ready-made one you wish to quote"
    def __init__(self, text):
        self.text = text
    def save(self, file):
        file.write(self.text + LINEEND)



class PDFCatalog(PDFObject):
    "requires RefPages and RefOutlines set"
    def __init__(self):
        self.template = LINEEND.join([
                        '<<',
                        '/Type /Catalog',
                        '/Pages %d 0 R',
                        '/Outlines %d 0 R',
                        '>>'
                        ])
    def save(self, file):
        file.write((self.template % (self.RefPages, self.RefOutlines) + LINEEND).encode('utf-8'))


class PDFInfo(PDFObject):
    """PDF documents can have basic information embedded, viewable from
    File | Document Info in Acrobat Reader.  If this is wrong, you get
    Postscript errors while printing, even though it does not print."""
    def __init__(self):
        self.title = "untitled"
        self.author = "anonymous"
        self.subject = "unspecified"

        now = time.localtime(time.time())
        self.datestr = '%04d%02d%02d%02d%02d%02d' % tuple(now[0:6])

    def save(self, file):
        file.write((LINEEND.join([
                "<</Title (%s)",
                "/Author (%s)",
                "/CreationDate (D:%s)",
                "/Producer (PDFgen)",
                "/Subject (%s)",
                ">>"
                ]) % (
    pdfutils._escape(self.title),
    pdfutils._escape(self.author),
    self.datestr,
    pdfutils._escape(self.subject)
    ) + LINEEND).encode('utf-8'))


class PDFOutline(PDFObject):
    "null outline, does nothing yet"
    def __init__(self):
        self.template = LINEEND.join([
                '<<',
                '/Type /Outlines',
                '/Count 0',
                '>>'])
    def save(self, file):
        file.write((self.template + LINEEND).encode('utf-8'))


class PDFPageCollection(PDFObject):
    "presumes PageList attribute set (list of integers)"
    def __init__(self):
        self.PageList = []

    def save(self, file):
        lines = [ '<<',
                '/Type /Pages',
                '/Count %d' % len(self.PageList),
                '/Kids ['
                ]
        for page in self.PageList:
            lines.append(str(page) + ' 0 R ')
        lines.append(']')
        lines.append('>>')
        text = LINEEND.join(lines)
        file.write((text + LINEEND).encode('utf-8'))


class PDFPage(PDFObject):
    """The Bastard.  Needs list of Resources etc. Use a standard one for now.
    It manages a PDFStream object which must be added to the document's list
    of objects as well."""
    def __init__(self):
        self.drawables = []
        self.pagewidth = 595  #these are overridden by piddlePDF
        self.pageheight = 842
        self.stream = PDFStream()
        self.hasImages = 0
        self.pageTransitionString = ''  # presentation effects
        # editors on different systems may put different things in the line end
        # without me noticing.  No triple-quoted strings allowed!
        self.template = LINEEND.join([
                '<<',
                '/Type /Page',
                '/Parent %(parentpos)d 0 R',
                '/Resources',
                '   <<',
                '   /Font %(fontdict)s',
                '   /ProcSet %(procsettext)s',
                '   >>',
                '/MediaBox [0 0 %(pagewidth)d %(pageheight)d]',  #A4 by default
                '/Contents %(contentspos)d 0 R',
                '%(transitionString)s',
                '>>'])

    def setCompression(self, onoff=0):
        "Turns page compression on or off"
        assert onoff in [0, 1], "Page compression options are 1=on, 2=off"
        self.stream.compression = onoff

    def save(self, file):
        self.info['pagewidth'] = self.pagewidth
        self.info['pageheight'] = self.pageheight
        # check for image support
        if self.hasImages:
            self.info['procsettext'] = '[/PDF /Text /ImageC]'
        else:
            self.info['procsettext'] = '[/PDF /Text]'
        self.info['transitionString'] = self.pageTransitionString

        file.write((self.template % self.info + LINEEND).encode('utf-8'))

    def clear(self):
        self.drawables = []

    def setStream(self, data):
        if isinstance(data, list):
            data = LINEEND.join(data)
        self.stream.setStream(data)

TestStream = "BT /F6 24 Tf 80 672 Td 24 TL (   ) Tj T* ET"


class PDFStream(PDFObject):
    "Used for the contents of a page"
    def __init__(self):
        self.data = None
        self.compression = 0

    def setStream(self, data):
        self.data = data

    def save(self, file):
        #avoid crashes if they wrote nothing in the page
        if self.data is None:
             self.data = TestStream

        if self.compression == 1:
            comp = zlib.compress((self.data).encode('utf-8'))   #this bit is very fast...
            base85 = pdfutils._AsciiBase85Encode(comp) #...sadly this isn't
            wrapped = pdfutils._wrap(base85)
            data_to_write = wrapped
        else:
            data_to_write = self.data
        # the PDF length key should contain the length including
        # any extra LF pairs added by Print on DOS.

        #lines = len(string.split(self.data,'\n'))
        #length = len(self.data) + lines   # one extra LF each
        length = len(data_to_write) + len(LINEEND)    #AR 19980202
        if self.compression:
            file.write(('<< /Length %d /Filter [/ASCII85Decode /FlateDecode]>>' % length + LINEEND).encode('utf-8'))
        else:
            file.write(('<< /Length %d >>' % length + LINEEND).encode('utf-8'))
        file.write(('stream' + LINEEND).encode('utf-8'))
        file.write((data_to_write + LINEEND).encode('latin-1'))  # Really?
        file.write(('endstream' + LINEEND).encode('utf-8'))

class PDFImage(PDFObject):
    # sample one while developing.  Currently, images go in a literals
    def save(self, file):
        file.write(LINEEND.join([
                '<<',
                '/Type /XObject',
                '/Subtype /Image',
                '/Name /Im0',
                '/Width 24',
                '/Height 23',
                '/BitsPerComponent 1',
                '/ColorSpace /DeviceGray',
                '/Filter /ASCIIHexDecode',
                '/Length 174',
                '>>',
                'stream',
                '003B00 002700 002480 0E4940 114920 14B220 3CB650',
                '75FE88 17FF8C 175F14 1C07E2 3803C4 703182 F8EDFC',
                'B2BBC2 BB6F84 31BFC2 18EA3C 0E3E00 07FC00 03F800',
                '1E1800 1FF800>',
                'endstream',
                'endobj'
                ]) + LINEEND)

class PDFEmbddedFont(PDFStream):
    def __init__(self, fontStreamType):
        PDFStream.__init__(self)
        self.fontType = fontStreamType

    def save(self, file):
        #avoid crashes if they wrote nothing in the page
        if self.data is None:
             self.data = TestStream

        if self.compression == 1:
            comp = zlib.compress(self.data)   #this bit is very fast...
            base85 = pdfutils._AsciiBase85Encode(comp) #...sadly this isn't
            data_to_write = pdfutils._wrap(base85)
        else:
            data_to_write = self.data
        # the PDF length key should contain the length including
        # any extra LF pairs added by Print on DOS.

        #lines = len(string.split(self.data,'\n'))
        #length = len(self.data) + lines   # one extra LF each
        length = len(data_to_write) + len(LINEEND)    #AR 19980202
        if self.fontType is None:
            fontStreamEntry = ""
        else:
            fontStreamEntry = "/Subtype %s" % (self.fontType)

        if self.compression:
            file.write(('<<  %s /Length %d /Filter [/ASCII85Decode /FlateDecode] >>' % (fontStreamEntry, length) + LINEEND).encode('utf-8'))
        else:
            file.write(('<< /Length %d %s >>' % (length,  fontStreamEntry) + LINEEND).encode('utf-8'))
        file.write(('stream' + LINEEND).encode('utf-8'))
        file.write(data_to_write + b'\r\n')
        file.write(('endstream' + LINEEND).encode('utf-8'))

class PDFType1Font(PDFObject):
    def __init__(self, key, psName, encoding=kDefaultEncoding, fontDescriptObjectNumber = None, type = kDefaultFontType, firstChar = None, lastChar = None, widths = None):
        self.fontname = psName
        self.keyname = key
        if encoding is None:
           encoding = '/MacRomanEncoding'
        textList = [
                    '<<',
                    '/Type /Font',
                    '/Subtype %s' % type]
        if fontDescriptObjectNumber != None:
            textList.append("/FontDescriptor %s 0 R" % fontDescriptObjectNumber)
            textList.append("/FirstChar %s" % firstChar)
            textList.append("/LastChar %s" % lastChar)
            textList.append("/Widths %s" % widths)
        textList += ['/Name /%s', # filled in in 'save'
                    '/BaseFont /%s',# filled in in 'save'
                    "/Encoding %s" % (encoding),
                    '>>']
        self.template = LINEEND.join(textList)

    def save(self, file):
        file.write((self.template % (self.keyname, self.fontname) + LINEEND).encode('utf-8'))


class PDFType0Font(PDFType1Font):
    pass

class PDFontDescriptor(PDFObject):
    def __init__(self, text):
        self.text = text

    def save(self, file):
        file.write((self.text + LINEEND).encode('utf-8'))





##############################################################
#
#            some helpers
#
##############################################################


def MakeFontDictionary(fontMapping):
    # fontMapping[psName] = [fontIndex, pdfFont, objIndex]
    dict_ = "  <<" + LINEEND
    entries = sorted(fontMapping.values())
    for entry in entries:
        objectPos = entry[2]
        pdfFont = entry[1]
        dict_ = dict_ + '\t\t/%s %d 0 R ' % (pdfFont.keyname, objectPos) + LINEEND
    dict_ = dict_ + "\t\t>>" + LINEEND
    return dict_



import os
from shutil import copytree
import pytest
import subprocess

from afdko.fdkutils import get_temp_dir_path
from test_utils import (
    get_input_path,
    get_expected_path,
    generate_ttx_dump,
)
from runner import main as runner
from differ import main as differ

TOOL = 'buildcff2vf'
CMD = ['-t', TOOL]


# -----
# Tests
# -----

def test_rvrn_vf():
    input_dir = get_input_path('GSUBVar')
    temp_dir = get_temp_dir_path('GSUBVar')
    copytree(input_dir, temp_dir)
    ds_path = os.path.join(temp_dir, 'GSUBVar.designspace')
    runner(CMD + ['-o', 'd', f'_{ds_path}'])
    actual_path = os.path.join(temp_dir, 'GSUBVar.otf')
    actual_ttx = generate_ttx_dump(actual_path,
                                   ['CFF2', 'GSUB', 'avar', 'fvar'])
    expected_ttx = get_expected_path('GSUBVar.ttx')
    assert differ([expected_ttx, actual_ttx, '-s', '<ttFont sfntVersion'])


def test_cjk_vf():
    input_dir = get_input_path('CJKVar')
    temp_dir = get_temp_dir_path('CJKVar')
    copytree(input_dir, temp_dir)
    ds_path = os.path.join(temp_dir, 'CJKVar.designspace')
    runner(CMD + ['-o', 'd', f'_{ds_path}'])
    actual_path = os.path.join(temp_dir, 'CJKVar.otf')
    actual_ttx = generate_ttx_dump(actual_path,
                                   ['CFF2', 'HVAR', 'avar', 'fvar'])
    expected_ttx = get_expected_path('CJKVar.ttx')
    assert differ([expected_ttx, actual_ttx, '-s', '<ttFont sfntVersion'])


def test_sparse_cjk_vf():
    # subset fonts to lists in 'SHSansJPVFTest.subset.txt'
    # Check compatibility. Issue is in cid00089.
    # keep PostScript names
    input_dir = get_input_path('CJKSparseVar')
    temp_dir = get_temp_dir_path('CJKSparseVar')
    copytree(input_dir, temp_dir)
    ds_path = os.path.join(temp_dir, 'SHSansJPVFTest.designspace')
    subset_path = os.path.join(temp_dir, 'SHSansJPVFTest.subset.txt')
    runner(CMD + ['-o', 'd', f'_{ds_path}', 'i', f'_{subset_path}', 'c', 'k'])
    actual_path = os.path.join(temp_dir, 'SHSansJPVFTest.otf')
    actual_ttx = generate_ttx_dump(actual_path,
                                   ['CFF2', 'HVAR', 'avar', 'fvar'])
    expected_ttx = get_expected_path('SHSansJPVFTest.ttx')
    assert differ([expected_ttx, actual_ttx, '-s', '<ttFont sfntVersion'])


def test_compatibility_vf():
    input_dir = get_input_path('bug816')
    temp_dir = get_temp_dir_path('bug816var')
    copytree(input_dir, temp_dir)
    ds_path = os.path.join(temp_dir, 'bug816.designspace')
    runner(CMD + ['-o', 'c', 'd', f'_{ds_path}'])
    actual_path = os.path.join(temp_dir, 'bug816.otf')
    actual_ttx = generate_ttx_dump(actual_path, ['CFF2'])
    expected_ttx = get_expected_path('bug816.ttx')
    assert differ([expected_ttx, actual_ttx, '-s', '<ttFont sfntVersion'])


def test_subset_vf():
    input_dir = get_input_path('bug817')
    temp_dir = get_temp_dir_path('bug817var')
    copytree(input_dir, temp_dir)
    ds_path = os.path.join(temp_dir, 'bug817.designspace')
    subset_path = os.path.join(temp_dir, 'bug817.subset')
    runner(CMD + ['-o', 'd', f'_{ds_path}', 'i', f'_{subset_path}'])
    actual_path = os.path.join(temp_dir, 'bug817.otf')
    actual_ttx = generate_ttx_dump(actual_path, ['GSUB'])
    expected_ttx = get_expected_path('bug817.ttx')
    assert differ([expected_ttx, actual_ttx, '-s', '<ttFont sfntVersion'])


def test_bug1003(capfd):
    """
    Check that "the input set requires compatibilization" message is in
    stderr when run without '-c' option.
    """
    input_dir = get_input_path('bug1003')
    temp_dir = get_temp_dir_path('bug1003var')
    copytree(input_dir, temp_dir)
    ds_path = os.path.join(temp_dir, 'bug1003.designspace')
    runner(CMD + ['-o', 'd', f'_{ds_path}'])
    captured = capfd.readouterr()
    assert "The input set requires compatibilization" in captured.err


def test_bug1003_compat():
    """
    Check that file is properly built when '-c' is specified.
    """
    input_dir = get_input_path('bug1003')
    temp_dir = get_temp_dir_path('bug1003cvar')
    copytree(input_dir, temp_dir)
    ds_path = os.path.join(temp_dir, 'bug1003.designspace')
    runner(CMD + ['-o', 'c' 'd', f'_{ds_path}'])
    actual_path = os.path.join(temp_dir, 'bug1003.otf')
    actual_ttx = generate_ttx_dump(actual_path, ['CFF2'])
    expected_ttx = get_expected_path('bug1003.ttx')
    assert differ([expected_ttx, actual_ttx, '-s', '<ttFont sfntVersion'])


def test_stat_axis_not_in_fvar():
    """
    Fail if fvar axes are missing from STAT
    """
    input_dir = get_input_path('STAT_tests')
    temp_dir = get_temp_dir_path('STAT_tests')
    copytree(input_dir, temp_dir)
    ds_path = os.path.join(temp_dir,
                           'STAT_axis_missing/STAT_axis_missing.designspace')
    with pytest.raises(subprocess.CalledProcessError):
        runner(CMD + ['-o', 'd', f'_{ds_path}'])


def test_stat_axis_values_range():
    """
    Fail if STAT axis values are outside range defined in fvar
    """
    input_dir = get_input_path('STAT_tests')
    temp_dir = get_temp_dir_path('STAT_tests')
    copytree(input_dir, temp_dir)
    ds_path = os.path.join(temp_dir,
                           'STAT_axis_ranges/STAT_axis_ranges.designspace')
    with pytest.raises(subprocess.CalledProcessError):
        runner(CMD + ['-o', 'd', f'_{ds_path}'])


def test_stat_axis_in_fvar():
    """
    Basic check that fvar axes are defined in STAT
    """
    input_dir = get_input_path('STAT_tests')
    temp_dir = get_temp_dir_path('STAT_tests')
    copytree(input_dir, temp_dir)
    ds_path = os.path.join(temp_dir,
                           'STAT_axis_in_fvar/STAT_axis_in_fvar.designspace')
    runner(CMD + ['-o', 'd', f'_{ds_path}'])
    actual_path = os.path.join(temp_dir,
                               'STAT_axis_in_fvar/STAT_axis_in_fvar.otf')
    actual_ttx = generate_ttx_dump(actual_path, ['STAT'])
    expected_ttx = get_expected_path('STAT_in_fvar.ttx')
    assert differ([expected_ttx, actual_ttx, '-s', '<ttFont sfntVersion'])


def test_stat_infinite_range():
    """
    Check that range tests work when infinite ranges are defined
    """
    input_dir = get_input_path('STAT_tests')
    temp_dir = get_temp_dir_path('STAT_tests')
    copytree(input_dir, temp_dir)
    ds_path = os.path.join(temp_dir,
                           'STAT_infinite_range/'
                           'STAT_infinite_range.designspace')
    runner(CMD + ['-o', 'd', f'_{ds_path}'])
    actual_path = os.path.join(temp_dir,
                               'STAT_infinite_range/STAT_infinite_range.otf')
    actual_ttx = generate_ttx_dump(actual_path, ['STAT'])
    expected_ttx = get_expected_path('STAT_infinite_range.ttx')
    assert differ([expected_ttx, actual_ttx, '-s', '<ttFont sfntVersion'])



import argparse
import os
import pytest
import subprocess

from runner import (
    main as runner,
    get_options,
    _check_tool,
    _check_save_path,
)

TESTS_DIR = os.path.split(__file__)[0]


def test_bad_tx_cmd():
    # Trigger a fatal error from a command line tool, to make sure
    # it is handled correctly.
    with pytest.raises(subprocess.CalledProcessError):
        runner(['-t', 'tx', '-o', 'bad_opt'])


def test_check_tool_error():
    assert isinstance(_check_tool('not_a_tool'), tuple)


def test_check_tool_error_cli():
    with pytest.raises(SystemExit) as exc_info:
        runner(['-t', 'not_a_tool'])
    assert exc_info.value.code == 2


@pytest.mark.parametrize('tool_name', [
    'detype1', 'type1', 'sfntedit', 'sfntdiff', 'makeotfexe'])
def test_check_tool_unhacked(tool_name):
    assert _check_tool(tool_name) == tool_name


def test_capture_error_message():
    output_path = runner(['-t', 'makeotfexe', '-s', '-e'])
    with open(output_path, 'rb') as f:
        output = f.read()
    assert b"[FATAL] Source font file not found: font.ps" in output


@pytest.mark.parametrize('v_arg', ['', 'v', 'vv', 'vvv'])
def test_log_level(v_arg):
    v_val = len(v_arg)
    arg = [] if not v_val else [f'-{v_arg}']
    opts = get_options(['-t', 'tx'] + arg)
    assert opts.verbose == v_val


def test_check_save_path__path_does_not_exist():
    tmp_fpath = os.path.join(TESTS_DIR, 'foobar')
    abs_tmp_fpath = os.path.abspath(os.path.realpath(tmp_fpath))
    assert os.path.exists(tmp_fpath) is False
    assert _check_save_path(tmp_fpath) == abs_tmp_fpath
    assert os.path.exists(tmp_fpath) is False


def test_check_save_path__path_exists():
    tmp_fpath = os.path.join(TESTS_DIR, 'runner_data', 'input', 'a_file')
    abs_tmp_fpath = os.path.abspath(os.path.realpath(tmp_fpath))
    assert os.path.exists(tmp_fpath) is True
    assert _check_save_path(tmp_fpath) == abs_tmp_fpath
    assert os.path.exists(tmp_fpath) is True


def test_check_save_path__path_is_folder():
    tmp_fpath = os.path.join(TESTS_DIR, 'runner_data', 'input')
    assert os.path.exists(tmp_fpath) is True
    with pytest.raises(argparse.ArgumentTypeError):
        _check_save_path(tmp_fpath)



# Copyright 2014 Adobe. All rights reserved.
"""
otfpdf v1.6.4 Nov 25 2019
provides support for the proofpdf script, for working with OpenType/CFF
fonts. Provides an implementation of the fontpdf font object. Cannot be
run alone.
"""
from fontTools.pens.boundsPen import BoundsPen
from fontTools.misc.psCharStrings import T2OutlineExtractor

from afdko.pdflib.fontpdf import FontPDFGlyph, FontPDFFont, FontPDFPen

__version__ = "1.6.4"


class txPDFFont(FontPDFFont):

    def __init__(self, clientFont, params):
        self.clientFont = clientFont
        if params.userBaseLine is not None:
            self.baseLine = params.userBaseLine
        else:
            self.baseLine = None
        self.path = params.rt_filePath
        self.isCID = 0
        self.psName = None
        self.OTVersion = None
        self.emSquare = None
        self.bbox = None
        self.ascent = None
        self.descent = None
        self.blueZones = None
        self.getEmSquare()
        self.getBaseLine()
        self.getBBox()
        self.GetBlueZones()
        self.AscentDescent()

    def __lt__(self, other):
        return self.clientGetPSName() < other.clientGetPSName()

    def clientGetPSName(self):
        psName = self.clientFont['CFF '].cff.fontNames[0]
        topDict = self.clientFont['CFF '].cff.topDictIndex[0]
        if hasattr(topDict, "ROS"):
            self.isCID = 1
        return psName

    def clientGetOTVersion(self):
        try:
            version = self.clientFont['head'].fontRevision
        except KeyError:
            # Maybe its a dummy OTF made from a Type 1 font,
            # and has only a CFF table.
            try:
                topDict = self.clientFont['CFF '].cff.topDictIndex[0]
                if hasattr(topDict, "ROS"):
                    version = topDict.CIDFontVersion  # a number
                else:
                    version = topDict.version  # a string
                    version = eval(version)
            except AttributeError:
                version = 1.0

        majorVersion = int(version)
        minorVersion = str(
            int(1000 * (0.0005 + version - majorVersion))).zfill(3)
        versionString = "%s.%s" % (majorVersion, minorVersion)
        return versionString

    def clientGetGlyph(self, glyphName):
        return txPDFGlyph(self, glyphName)

    def clientGetEmSquare(self):
        try:
            emSquare = self.clientFont['head'].unitsPerEm
        except KeyError:
            # Maybe its a dummy OTF made from a Type 1 font,
            # and has only a CFF table.
            topDict = self.clientFont['CFF '].cff.topDictIndex[0]
            scale = topDict.FontMatrix[0]
            emSquare = int(0.5 + 1 / scale)
        return emSquare

    def clientGetBaseline(self):
        baseLine = 0
        txFont = self.clientFont
        try:
            unicodeRange2 = txFont["OS/2"].ulUnicodeRange2
            if unicodeRange2 & 0x10000:  # supports CJK ideographs
                baseTag = "ideo"
            else:
                baseTag = "romn"
            horiz_axis = self.clientFont['BASE'].table.HorizAxis
            baseTagIndex = horiz_axis.BaseTagList.BaselineTag.index(baseTag)
            for baseRecord in horiz_axis.BaseScriptList.BaseScriptRecord:
                if baseRecord.BaseScriptTag == "latn":
                    baseValues = baseRecord.BaseScript.BaseValues
                    baseLine = baseValues.BaseCoord[baseTagIndex].Coordinate
                    break
        except (KeyError, AttributeError):
            topDict = txFont['CFF '].cff.topDictIndex[0]
            if hasattr(topDict, "ROS"):
                baseLine = -120
            else:
                baseLine = 0

        return baseLine

    def clientGetBBox(self):
        try:
            headTable = self.clientFont['head']
            return [headTable.xMin, headTable.yMin,
                    headTable.xMax, headTable.yMax]
        except KeyError:
            # Maybe it's a dummy OTF made from a Type 1 font,
            # and has only a CFF table.
            topDict = self.clientFont['CFF '].cff.topDictIndex[0]
            bbox = topDict.FontBBox
            return [bbox[0], bbox[1], bbox[2], bbox[3]]

    def clientGetBlueZones(self):
        blueValues = []
        txFont = self.clientFont
        topDict = txFont['CFF '].cff.topDictIndex[0]
        if hasattr(topDict, "ROS"):
            FDArray = topDict.FDArray
            self.isCID = 1
        else:
            FDArray = [topDict]
        for fontDict in FDArray:
            try:
                rawBlueList = fontDict.Private.BlueValues
                blueList = []
                for i in list(range(0, len(rawBlueList), 2)):
                    blueList.append((rawBlueList[i], rawBlueList[i + 1]))
                blueValues.append(blueList)
            except AttributeError:
                # the font has no BlueValues
                pass
        return blueValues

    def clientGetAscentDescent(self):
        try:
            os2Table = self.clientFont['OS/2']
            return os2Table.sTypoAscender, os2Table.sTypoDescender
        except KeyError:
            return None, None


def hintOn(i, hintMaskBytes):
    # used to add the active hints to the bez string,
    # when a T2 hintmask operator is encountered.
    byteIndex = i // 8
    byteValue = hintMaskBytes[byteIndex]
    offset = 7 - (i % 8)
    return ((2 ** offset) & byteValue) > 0


class FontPDFT2OutlineExtractor(T2OutlineExtractor):

    def __init__(self, pen, localSubrs, globalSubrs, nominalWidthX,
                 defaultWidthX):
        T2OutlineExtractor.__init__(self, pen, localSubrs, globalSubrs,
                                    nominalWidthX, defaultWidthX)
        self.hintTable = []
        self.vhints = []
        self.hhints = []

    def op_hstem(self, index):
        args = self.popallWidth()
        self.updateHints(args, self.hhints)

    def op_vstem(self, index):
        args = self.popallWidth()
        self.updateHints(args, self.vhints)

    def op_hstemhm(self, index):
        args = self.popallWidth()
        self.hhints = []
        self.updateHints(args, self.hhints)

    def op_vstemhm(self, index):
        args = self.popallWidth()
        self.vhints = []
        self.updateHints(args, self.vhints)

    def op_hintmask(self, index):
        hintMaskString, index = self.doMask(index, "hintmask")
        self.firstOpSeen = 1
        return hintMaskString, index

    def op_cntrmask(self, index):
        hintMaskString, index = self.doMask(index, "cntrmask")
        self.firstOpSeen = 1
        return hintMaskString, index

    def updateHints(self, args, hintList):
        self.countHints(args)

        # first hint value is absolute hint coordinate, second is hint width
        lastval = args[0]
        arg = str(lastval)
        hint1 = arg

        for i in list(range(len(args))[1:]):
            val = args[i]
            newVal = lastval + val
            lastval = newVal

            if i % 2:
                hint2 = str(val)
                hintList.append((hint1, hint2))
            else:
                hint1 = str(newVal)

    def getCurHints(self, hintMaskBytes):
        curhhints = []
        curvhints = []
        numhhints = len(self.hhints)

        for i in list(range(numhhints)):
            if hintOn(i, hintMaskBytes):
                curhhints.append(i)
        numvhints = len(self.vhints)
        for i in list(range(numvhints)):
            if hintOn(i + numhhints, hintMaskBytes):
                curvhints.append(i)
        return curhhints, curvhints

    def doMask(self, index, maskCommand):
        if not self.hintMaskBytes:
            args = self.popallWidth()
            if args:
                self.vhints = []
                self.updateHints(args, self.vhints)
            self.hintMaskBytes = (self.hintCount + 7) // 8

        self.hintMaskString, index = (
            self.callingStack[-1].getBytes(index, self.hintMaskBytes))

        curhhints, curvhints = self.getCurHints(self.hintMaskString)

        self.hintTable.append([curhhints, curvhints, self.pen.total])

        return self.hintMaskString, index

    def countHints(self, args):
        self.hintCount = self.hintCount + len(args) // 2


def drawCharString(charString, pen):
    subrs = getattr(charString.private, "Subrs", [])
    extractor = FontPDFT2OutlineExtractor(
        pen, subrs, charString.globalSubrs, charString.private.nominalWidthX,
        charString.private.defaultWidthX)
    extractor.execute(charString)
    charString.width = extractor.width
    charString.hintTable = extractor.hintTable
    charString.hhints = extractor.hhints
    charString.vhints = extractor.vhints


class txPDFGlyph(FontPDFGlyph):

    def clientInitData(self):
        txFont = self.parentFont.clientFont

        try:
            glyph_set = txFont.getGlyphSet()
        except KeyError:
            glyph_set = None

        if not hasattr(txFont, 'vmetrics'):
            try:
                txFont.vmetrics = txFont['vmtx'].metrics
            except KeyError:
                txFont.vmetrics = None
            try:
                txFont.vorg = txFont['VORG']
            except KeyError:
                txFont.vorg = None

        fTopDict = txFont['CFF '].cff.topDictIndex[0]
        self.isCID = self.parentFont.isCID
        charstring = fTopDict.CharStrings[self.name]

        # Get the list of points
        pen = FontPDFPen(glyph_set)
        drawCharString(charstring, pen)
        self.hintTable = charstring.hintTable

        self.hhints = charstring.hhints
        self.vhints = charstring.vhints
        self.numMT = pen.numMT
        self.numLT = pen.numLT
        self.numCT = pen.numCT
        self.numPaths = pen.numPaths
        self.pathList = pen.pathList
        for path in self.pathList:
            lenPath = len(path)
            path[-1].next = path[0]
            path[0].last = path[-1]
            if lenPath > 1:
                path[0].next = path[1]
                path[-1].last = path[-2]
                for i in list(range(lenPath)[1:-1]):
                    pt = path[i]
                    pt.next = path[i + 1]
                    pt.last = path[i - 1]

        assert len(self.pathList) == self.numPaths, (
            "Path lengths don't match %s %s" % (len(self.pathList),
                                                self.numPaths))
        # get the bbox and width.
        pen = BoundsPen(glyph_set)
        charstring.draw(pen)
        self.xAdvance = charstring.width
        glyph_bounds = pen.bounds
        if not glyph_bounds:
            self.BBox = [0, 0, 0, 0]
        else:
            self.BBox = [round(item) for item in glyph_bounds]

        self.yOrigin = self.parentFont.emSquare + self.parentFont.getBaseLine()
        if txFont.vorg:
            try:
                self.yOrigin = txFont.vorg[self.name]
            except KeyError:
                if txFont.vmetrics:
                    try:
                        mtx = txFont.vmetrics[self.name]
                        self.yOrigin = mtx[1] + self.BBox[3]
                    except KeyError:
                        pass

        haveVMTX = 0
        if txFont.vmetrics:
            try:
                mtx = txFont.vmetrics[self.name]
                self.yAdvance = mtx[0]
                self.tsb = mtx[1]
                haveVMTX = 1
            except KeyError:
                pass
        if not haveVMTX:
            self.yAdvance = self.parentFont.getEmSquare()
            self.tsb = (
                self.yOrigin - self.BBox[3] + self.parentFont.getBaseLine())

        # Get the fdIndex, so we can later determine
        # which set of blue values to use.
        self.fdIndex = 0
        if hasattr(fTopDict, "ROS"):
            gid = fTopDict.CharStrings.charStrings[self.name]
            self.fdIndex = fTopDict.FDSelect[gid]



import os
import pytest
import subprocess

from runner import main as runner
from differ import main as differ
from test_utils import get_input_path, get_expected_path, get_temp_file_path

TOOL = 'comparefamily'
CMD = ['-t', TOOL]


# -----
# Tests
# -----

@pytest.mark.parametrize('font_format', ['otf', 'ttf'])
@pytest.mark.parametrize('font_family', ['source-code-pro'])
def test_report(font_family, font_format):
    input_dir = os.path.join(get_input_path(font_family), font_format)
    log_path = get_temp_file_path()
    runner(CMD + ['-o', 'd', f'_{input_dir}', 'tolerance', '_3',
                  'rm', 'rn', 'rp', 'l', f'_{log_path}'])
    expected_path = get_expected_path(f'{font_family}_{font_format}.txt')
    assert differ([expected_path, log_path, '-l', '1'])


def test_report2():
    input_dir = get_input_path('font-family')
    expected_path = get_expected_path('font-family.txt')
    log_path = get_temp_file_path()
    runner(CMD + ['-o', 'd', f'_{input_dir}',
                  'rm', 'rn', 'rp', 'l', f'_{log_path}'])
    assert differ([expected_path, log_path, '-l', '1'])


def test_st28_basic_cmap():
    input_dir = get_input_path('basic_cmap')
    expected_path = get_expected_path('st28_basic_cmap.txt')
    log_path = get_temp_file_path()
    runner(CMD + ['-o', 'st', '_28', 'd', f'_{input_dir}',
                  'l', f'_{log_path}'])
    assert differ([expected_path, log_path, '-l', '1'])


def test_font_with_no_cmap():
    # A tx fatal error happens: "OTF: can't find cmap"
    input_dir = get_input_path('no_cmap')
    with pytest.raises(subprocess.CalledProcessError) as err:
        runner(CMD + ['-o', 'd', f'_{input_dir}'])
    assert err.value.returncode == 1



import pytest
import subprocess

from runner import main as runner
from differ import main as differ
from test_utils import (get_input_path, get_expected_path, get_temp_file_path,
                        generate_ttx_dump, font_has_table)

TOOL = 'sfntedit'
CMD = ['-t', TOOL]

LIGHT = 'light.otf'
ITALIC = 'italic.otf'


# -----
# Tests
# -----

def test_exit_no_option():
    # When ran by itself, 'sfntedit' prints the usage
    assert subprocess.call([TOOL]) == 1


@pytest.mark.parametrize('arg', ['-h', '-u'])
def test_exit_known_option(arg):
    assert subprocess.call([TOOL, arg]) == 0


@pytest.mark.parametrize('arg', ['-j', '--bogus'])
def test_exit_unknown_option(arg):
    assert subprocess.call([TOOL, arg]) == 1


def test_no_options():
    actual_path = runner(CMD + ['-s', '-f', LIGHT])
    expected_path = get_expected_path('list_sfnt.txt')
    assert differ([expected_path, actual_path, '-l', '1'])


def test_extract_table():
    actual_path = get_temp_file_path()
    runner(CMD + ['-o', 'x', f'_GDEF={actual_path}', '-f', LIGHT])
    expected_path = get_expected_path('GDEF_light.tb')
    assert differ([expected_path, actual_path, '-m', 'bin'])

# This test fails on Windows and passes on Mac
# https://ci.appveyor.com/project/adobe-type-tools/afdko/build/1.0.108
# https://travis-ci.org/adobe-type-tools/afdko/builds/372889493
# def test_delete_table():
#     actual_path = runner(CMD + ['-o', 'd', '_GDEF', '-f', LIGHT])
#     expected_path = get_expected_path('light_n_GDEF.otf')
#     assert differ([expected_path, actual_path, '-m', 'bin'])


def test_add_table():
    table_path = get_input_path('GDEF_italic.tb')
    font_path = get_input_path(ITALIC)
    actual_path = get_temp_file_path()
    assert font_has_table(font_path, 'GDEF') is False
    runner(CMD + ['-a', '-o', 'a', f'_GDEF={table_path}',
                  '-f', font_path, actual_path])
    expected_path = get_expected_path('italic_w_GDEF.otf')
    assert differ([expected_path, actual_path, '-m', 'bin']) is False
    assert font_has_table(actual_path, 'GDEF')
    actual_ttx = generate_ttx_dump(actual_path)
    expected_ttx = generate_ttx_dump(expected_path)
    assert differ([expected_ttx, actual_ttx, '-s', '    <checkSumAdjustment'])


def test_linux_ci_failure_bug570():
    table_path = get_input_path('1_fdict.cff')
    font_path = get_input_path('core.otf')
    actual_path = get_temp_file_path()
    runner(CMD + ['-a', '-o', 'a', f'_CFF={table_path}',
                  '-f', font_path, actual_path])
    expected_path = get_expected_path('1_fdict.otf')
    assert differ([expected_path, actual_path, '-m', 'bin'])


def test_missing_table_delete_bug160():
    actual_path = get_temp_file_path()
    stderr_path = runner(CMD + ['-s', '-e', '-o', 'd', '_xyz,GSUB',
                                '-f', LIGHT, actual_path])
    with open(stderr_path, 'rb') as f:
        output = f.read()
    assert b'[WARNING]: table missing (xyz )' in output
    assert font_has_table(get_input_path(LIGHT), 'GSUB')
    assert not font_has_table(actual_path, 'GSUB')


def test_missing_table_extract_bug160():
    actual_path = get_temp_file_path()
    stderr_path = runner(CMD + ['-s', '-e', '-f', LIGHT, actual_path, '-o',
                                'x', f'_xyz,head={actual_path}'])
    with open(stderr_path, 'rb') as f:
        output = f.read()
    assert b'[WARNING]: table missing (xyz )' in output
    expected_path = get_expected_path('head_light.tb')
    assert differ([expected_path, actual_path, '-m', 'bin'])



import argparse
import glob
import logging
import os
import sys
from functools import partial, singledispatch
from itertools import chain
from multiprocessing import Pool

from fontTools import configLogger
from fontTools.misc.cliTools import makeOutputFileName
from fontTools.pens.cu2quPen import Cu2QuPen
from fontTools.pens.ttGlyphPen import TTGlyphPen
from fontTools.ttLib import TTCollection, TTFont, TTLibError, newTable

log = logging.getLogger()
configLogger(logger=log)

# default approximation error, measured in UPEM
MAX_ERR = 1.0

# default 'post' table format
POST_FORMAT = 2.0

# assuming the input contours' direction is correctly set (counter-clockwise),
# we just flip it to clockwise
REVERSE_DIRECTION = True


def glyphs_to_quadratic(
        glyphs, max_err=MAX_ERR, reverse_direction=REVERSE_DIRECTION):
    quadGlyphs = {}
    for gname in glyphs.keys():
        glyph = glyphs[gname]
        ttPen = TTGlyphPen(glyphs)
        cu2quPen = Cu2QuPen(ttPen, max_err,
                            reverse_direction=reverse_direction)
        glyph.draw(cu2quPen)
        quadGlyphs[gname] = ttPen.glyph()
    return quadGlyphs


def update_hmtx(ttFont, glyf):
    hmtx = ttFont["hmtx"]
    for glyphName, glyph in glyf.glyphs.items():
        if hasattr(glyph, 'xMin'):
            hmtx[glyphName] = (hmtx[glyphName][0], glyph.xMin)


@singledispatch
def otf_to_ttf(ttFont, post_format=POST_FORMAT, **kwargs):
    if ttFont.sfntVersion != "OTTO":
        raise TTLibError("Not a OpenType font (bad sfntVersion)")
    assert "CFF " in ttFont

    glyphOrder = ttFont.getGlyphOrder()

    ttFont["loca"] = newTable("loca")
    ttFont["glyf"] = glyf = newTable("glyf")
    glyf.glyphOrder = glyphOrder
    glyf.glyphs = glyphs_to_quadratic(ttFont.getGlyphSet(), **kwargs)
    del ttFont["CFF "]
    if "VORG" in ttFont:
        del ttFont["VORG"]
    glyf.compile(ttFont)
    update_hmtx(ttFont, glyf)

    ttFont["maxp"] = maxp = newTable("maxp")
    maxp.tableVersion = 0x00010000
    maxp.maxZones = 1
    maxp.maxTwilightPoints = 0
    maxp.maxStorage = 0
    maxp.maxFunctionDefs = 0
    maxp.maxInstructionDefs = 0
    maxp.maxStackElements = 0
    maxp.maxSizeOfInstructions = 0
    maxp.maxComponentElements = max(
        len(g.components if hasattr(g, 'components') else [])
        for g in glyf.glyphs.values())
    maxp.compile(ttFont)

    post = ttFont["post"]
    post.formatType = post_format
    post.extraNames = []
    post.mapping = {}
    post.glyphOrder = glyphOrder
    try:
        post.compile(ttFont)
    except OverflowError:
        post.formatType = 3
        log.warning("Dropping glyph names, they do not fit in 'post' table.")

    ttFont.sfntVersion = "\000\001\000\000"


@otf_to_ttf.register(TTCollection)
def _(fonts, **kwargs):
    skip = 0
    for font in fonts:
        try:
            otf_to_ttf(font, **kwargs)
        except TTLibError as warn:
            skip += 1
            log.warning(warn)

    if skip == len(fonts):
        raise TTLibError("a Font Collection that has Not a OpenType font")


def run(path, options):
    try:
        font = TTFont(path, fontNumber=options.face_index)
        extension = '.ttf'
    except TTLibError:
        font = TTCollection(path)
        extension = '.ttc'

    if options.output and not os.path.isdir(options.output):
        output = options.output
    else:
        output = makeOutputFileName(path, outputDir=options.output,
                                    extension=extension,
                                    overWrite=options.overwrite)

    try:
        otf_to_ttf(font,
                   post_format=options.post_format,
                   max_err=options.max_error,
                   reverse_direction=options.reverse_direction)
    except TTLibError as warn:
        log.warning(f'"{path}" cannot be converted since it is {warn}.')
    else:
        font.save(output)


def main(args=None):
    parser = argparse.ArgumentParser()
    parser.add_argument("input", nargs='+', metavar="INPUT")
    parser.add_argument("-o", "--output")
    parser.add_argument("-e", "--max-error", type=float, default=MAX_ERR)
    parser.add_argument("--post-format", type=float, default=POST_FORMAT)
    parser.add_argument(
        "--keep-direction", dest='reverse_direction', action='store_false')
    parser.add_argument("--face-index", type=int, default=-1)
    parser.add_argument("--overwrite", action='store_true')
    options = parser.parse_args(args)

    if options.output and len(options.input) > 1:
        if not os.path.isdir(options.output):
            parser.error("-o/--output option must be a directory when "
                         "processing multiple fonts")

    files = chain.from_iterable(map(glob.glob, options.input))

    # Do not use "with" statement, or code coverage will malfunction.
    pool = Pool()
    try:
        pool.map(partial(run, options=options), files)
    finally:
        pool.close()
        pool.join()


if __name__ == "__main__":
    sys.exit(main())



import os
import pytest
from shutil import copy2, rmtree

from afdko.convertfonttocid import mergeFontToCFF

from differ import main as differ
from test_utils import (get_input_path, get_expected_path, get_temp_file_path,
                        generate_ttx_dump)

MODULE = 'convertfonttocid'

DATA_DIR = os.path.join(os.path.dirname(__file__), MODULE + '_data')
TEMP_DIR = os.path.join(DATA_DIR, 'temp_output')


def setup_module():
    """
    Create the temporary output directory
    """
    rmtree(TEMP_DIR, ignore_errors=True)
    os.mkdir(TEMP_DIR)


def teardown_module():
    """
    teardown the temporary output directory
    """
    rmtree(TEMP_DIR)


# -----
# Tests
# -----

@pytest.mark.parametrize('subroutinize', [True, False])
@pytest.mark.parametrize('font_filename', ['1_fdict.ps', '3_fdict.ps'])
def test_mergeFontToCFF_bug570(font_filename, subroutinize):
    # 'dir=TEMP_DIR' is used for guaranteeing that the temp data is on same
    # file system as other data; if it's not, a file rename step made by
    # sfntedit will NOT work.
    actual_path = get_temp_file_path(directory=TEMP_DIR)
    subr_str = 'subr' if subroutinize else 'no_subr'
    font_base = os.path.splitext(font_filename)[0]
    ttx_filename = f'{font_base}-{subr_str}.ttx'
    source_path = get_input_path(font_filename)
    output_path = get_input_path('core.otf')
    copy2(output_path, actual_path)

    mergeFontToCFF(source_path, actual_path, subroutinize)

    actual_ttx = generate_ttx_dump(actual_path, ['CFF '])
    expected_ttx = get_expected_path(ttx_filename)
    assert differ([expected_ttx, actual_ttx, '-l', '2'])



import inspect
import os
import shutil
import subprocess

from fontTools.ttLib import TTCollection, TTFont, TTLibError
from afdko.fdkutils import (get_temp_file_path, run_shell_command)

TIMEOUT = 240  # seconds


def generalizeCFF(otfPath, do_sfnt=True):
    """
    Adapted from similar routine in 'buildmasterotfs'. This uses
    temp files for both tx output and sfntedit output instead of
    overwriting 'otfPath', and also provides an option to skip
    the sfntedit step (so 'otfPath' can be either a .otf file or
    a .cff file).
    """
    tmp_tx_path = get_temp_file_path()
    out_path = get_temp_file_path()
    shutil.copyfile(otfPath, out_path, follow_symlinks=True)

    if not run_shell_command(['tx', '-cff', '+b', '-std', '-no_opt',
                              otfPath, tmp_tx_path]):
        raise Exception

    if do_sfnt:

        if not run_shell_command(['sfntedit', '-a',
                                  f'CFF ={tmp_tx_path}', out_path]):
            raise Exception

        return out_path

    else:

        return tmp_tx_path


def get_data_dir():
    tests_dir, test_file = os.path.split(inspect.stack()[2][1])
    tool_name = test_file[:-8]  # trim '_test.py' portion
    return os.path.join(tests_dir, tool_name + '_data')


def get_expected_path(file_name):
    return os.path.join(get_data_dir(), 'expected_output', file_name)


def get_input_path(file_name):
    return os.path.join(get_data_dir(), 'input', file_name)


def get_bad_input_path(file_name):
    return os.path.join(get_data_dir(), 'input', 'bad', file_name)


def generate_ttx_dump(font_path, tables=None):
    try:
        font = TTFont(font_path)
    except TTLibError:
        font = TTCollection(font_path)

    with font:
        temp_path = get_temp_file_path()
        font.saveXML(temp_path, tables=tables)
        return temp_path


def generate_spot_dumptables(font_path, tables):
    tmp_txt_path = get_temp_file_path()
    myargs = ['spot', "-t" + ",".join(tables), font_path]
    spot_txt = subprocess.check_output(myargs, timeout=TIMEOUT)
    tf = open(tmp_txt_path, "wb")
    tf.write(spot_txt)
    tf.close()
    return tmp_txt_path


def generate_ps_dump(font_path):
    with open(font_path, 'rb') as ps_file:
        data = ps_file.read()
    line = ''
    i = 0
    in_binary_section = False
    bytes_read = 0
    byte_count = 0
    temp_path = get_temp_file_path()
    with open(temp_path, 'w') as temp_file:
        while i < len(data):
            x = data[i]
            if not in_binary_section:
                if x != 0x0A:
                    line += '%c' % x
                else:
                    if line.startswith('%%BeginData'):
                        in_binary_section = True
                        byte_count = int(line.split()[1])
                    line += '\n'
                    temp_file.write(line)
                    line = ''
            else:
                bytes_read += 1
                if bytes_read == byte_count:
                    in_binary_section = False
                    if bytes_read % 32 != 1:
                        temp_file.write('\n')
                else:
                    temp_file.write('%02X' % x)
                    if bytes_read % 32 == 0:
                        temp_file.write('\n')
            i += 1
    return temp_path


def font_has_table(font_path, table_tag):
    with TTFont(font_path) as font:
        return table_tag in font


def get_font_revision(font_path):
    with TTFont(font_path) as font:
        return font['head'].fontRevision



import pytest
from shutil import copy2
import subprocess

from fontTools.ttLib import TTFont
from afdko import ttfdecomponentizer as ttfdecomp

from afdko.fdkutils import (
    get_temp_file_path,
    get_temp_dir_path,
)
from test_utils import (
    get_input_path,
    get_expected_path,
    generate_ttx_dump,
)
from runner import main as runner
from differ import main as differ

TOOL = 'ttfdecomponentizer'
CMD = ['-t', TOOL]

TEST_TTF_FILENAME = 'hascomponents.ttf'


def _get_test_ttf_path():
    return get_input_path(TEST_TTF_FILENAME)


def _read_file(file_path):
    with open(file_path, "rb") as f:
        return f.read()


def _write_file(file_path, data):
    with open(file_path, "wb") as f:
        f.write(data)


# -----
# Tests
# -----

def test_run_no_args():
    with pytest.raises(SystemExit) as exc_info:
        ttfdecomp.main()
    assert exc_info.value.code == 2


def test_run_cli_no_args():
    with pytest.raises(subprocess.CalledProcessError) as exc_info:
        runner(CMD)
    assert exc_info.value.returncode == 2


def test_run_invalid_font_path():
    with pytest.raises(SystemExit) as exc_info:
        ttfdecomp.main(['not_a_file'])
    assert exc_info.value.code == 2


def test_run_invalid_font_file():
    with pytest.raises(SystemExit) as exc_info:
        ttfdecomp.main(['not_a_font.ttf'])
    assert exc_info.value.code == 2


def test_run_with_output_path():
    ttf_path = _get_test_ttf_path()
    save_path = get_temp_file_path()
    ttfdecomp.main(['-o', save_path, ttf_path])
    font = TTFont(save_path)
    gtable = font['glyf']
    composites = [gtable[gl].isComposite() for gl in font.glyphOrder]
    assert not any(composites)


def test_run_cli_with_output_path():
    actual_path = get_temp_file_path()
    runner(CMD + ['-o', 'o', f'_{actual_path}',
                  f'_{get_input_path(TEST_TTF_FILENAME)}'])
    actual_ttx = generate_ttx_dump(actual_path, ['maxp', 'glyf'])
    expected_ttx = get_expected_path('nocomponents.ttx')
    assert differ([expected_ttx, actual_ttx, '-s', '<ttFont sfntVersion'])


def test_run_without_output_path():
    ttf_path = _get_test_ttf_path()
    temp_dir = get_temp_dir_path()
    save_path = get_temp_file_path(directory=temp_dir)
    copy2(ttf_path, save_path)
    ttfdecomp.main([save_path])
    font = TTFont(save_path)
    gtable = font['glyf']

    composites = [gtable[gl].isComposite() for gl in font.glyphOrder]
    assert not any(composites)


def test_options_help():
    with pytest.raises(SystemExit) as exc_info:
        ttfdecomp.main(['-h'])
    assert exc_info.value.code == 0


def test_options_version():
    with pytest.raises(SystemExit) as exc_info:
        ttfdecomp.main(['--version'])
    assert exc_info.value.code == 0


def test_options_bogus_option():
    with pytest.raises(SystemExit) as exc_info:
        ttfdecomp.main(['--bogus'])
    assert exc_info.value.code == 2


def test_options_dir_input():
    input_dir = get_temp_dir_path()
    font_path = _get_test_ttf_path()
    in_font_path = copy2(font_path, input_dir)
    ttfdecomp.main(['-d', input_dir, '-v'])
    font = TTFont(in_font_path)
    assert font['maxp'].maxComponentElements == 0


def test_options_dir_input_with_dir_output():
    input_dir = get_temp_dir_path()
    output_dir = get_temp_dir_path()
    font_path = _get_test_ttf_path()
    in_font_path = copy2(font_path, input_dir)
    ttfdecomp.main(['-d', input_dir, '-o', output_dir, '-v'])
    out_file = in_font_path.replace(input_dir, output_dir)
    font = TTFont(out_file)
    assert font['maxp'].maxComponentElements == 0


def test_options_bad_dir_input():
    with pytest.raises(SystemExit) as exc_info:
        input_dir = _get_test_ttf_path()
        ttfdecomp.main(['-d', input_dir])
    assert exc_info.value.code == 2


def test_options_bad_dir_output():
    with pytest.raises(SystemExit) as exc_info:
        input_dir = get_temp_dir_path()
        output_dir = _get_test_ttf_path()
        ttfdecomp.main(['-d', input_dir, '-o', output_dir])
    assert exc_info.value.code == 2


def test_options_good_dir_output_no_dir_input():
    with pytest.raises(SystemExit) as exc_info:
        ttf_path = _get_test_ttf_path()
        output_dir = get_temp_dir_path()
        ttfdecomp.main(['-o', output_dir, ttf_path])
    assert exc_info.value.code == 2



# Copyright 2016 Adobe. All rights reserved.

"""
A collection of functions used by multiple AFDKO scripts.
"""

import argparse
import os
import subprocess
import tempfile
import zipfile

__version__ = '1.3.7'

LEN_CID_TOK = 31  # length of string '%!PS-Adobe-3.0 Resource-CIDFont'


def validate_path(path_str):
    valid_path = os.path.abspath(os.path.realpath(path_str))
    if not os.path.exists(valid_path):
        raise argparse.ArgumentTypeError(
            f"'{path_str}' is not a valid path.")
    return valid_path


def get_temp_file_path(directory=None):
    file_descriptor, path = tempfile.mkstemp(dir=directory)
    os.close(file_descriptor)
    return path


def get_temp_dir_path(path_comp=None):
    path = tempfile.mkdtemp()
    if path_comp:
        path = os.path.join(path, path_comp)
    return path


def get_resources_dir():
    return os.path.join(os.path.dirname(__file__), 'resources')


def _font_is_ufo(path):
    from fontTools.ufoLib import UFOReader
    from fontTools.ufoLib.errors import UFOLibError
    try:
        UFOReader(path, validate=False)
        return True
    except (UFOLibError, KeyError, TypeError):
        return False


def get_font_format(font_file_path):
    if os.path.isdir(font_file_path):
        if _font_is_ufo(font_file_path):
            return 'UFO'
    elif os.path.isfile(font_file_path):
        with open(font_file_path, 'rb') as f:
            fullhead = f.read(LEN_CID_TOK)
            head = fullhead[0: 4]
            shorthead = fullhead[0: 2]
            if head == b'OTTO':
                return 'OTF'
            elif head in (b'\x00\x01\x00\x00', b'true'):
                return 'TTF'
            elif head == b'ttcf':
                return 'TTC'
            elif shorthead == b'\x01\x00':
                return 'CFF'
            elif shorthead == b'\x80\x01':
                return 'PFB'
            elif head in (b'%!PS', b'%!Fo'):
                if b'CID' not in fullhead:
                    return 'PFA'
                else:
                    return 'PFC'
            elif (shorthead == b'PK' and
                  zipfile.is_zipfile(font_file_path) and
                  _font_is_ufo(font_file_path)):
                return 'UFO'
    return None


def run_shell_command(args, suppress_output=False):
    """
    Runs a shell command.
    Returns True if the command was successful, and False otherwise.
    """
    sup = subprocess.DEVNULL if suppress_output else None

    try:
        subprocess.check_call(args, stderr=sup, stdout=sup)
        return True
    except (subprocess.CalledProcessError, OSError) as err:
        print(err)
        return False


def run_shell_command_logging(args):
    """
    Runs a shell command while logging both standard output and standard error.

    An earlier version of this function that used Popen.stdout.readline()
    was failing intermittently in CI, so now we're trying this version that
    uses Popen.communicate().
    """
    try:
        proc = subprocess.Popen(args,
                                stdout=subprocess.PIPE,
                                stderr=subprocess.STDOUT,
                                universal_newlines=True)
        out = proc.communicate()[0]
        print(out, end='')
        if proc.returncode != 0:
            msg = " ".join(args)
            print(f"Error executing command '{msg}'")
            return False
        else:
            return True
    except (subprocess.CalledProcessError, OSError) as err:
        msg = " ".join(args)
        print(f"Error executing command '{msg}'\n{err}")
        return False


def get_shell_command_output(args, std_error=False):
    """
    Runs a shell command and captures its output.
    To also capture standard error call with std_error=True.

    Returns a tuple; the first element will be True if the command was
    successful, and False otherwise. The second element will be a
    Unicode-encoded string, or None.
    """
    stderr = subprocess.STDOUT if std_error else None

    try:
        bytes_output = subprocess.check_output(args, stderr=stderr)
        str_output = bytes_output.decode('utf-8', 'backslashreplace')
        success = True

    except (subprocess.CalledProcessError, OSError) as err:
        print(err)
        success = False
        str_output = None

    return success, str_output


def runShellCmd(cmd, shell=True, timeout=None):
    try:
        p = subprocess.Popen(cmd, shell=shell, stdout=subprocess.PIPE,
                             stderr=subprocess.STDOUT)
        stdoutdata, _ = p.communicate(timeout=timeout)
        str_output = stdoutdata.decode('utf-8', 'backslashreplace')

        if p.returncode != 0:  # must be called *after* communicate()
            print(f"Error executing command '{cmd}'\n{str_output}")
            str_output = ""

    except subprocess.TimeoutExpired as err:
        p.kill()
        print(f"{err}")  # the 'err' will contain the command
        str_output = ""

    except (subprocess.CalledProcessError, OSError) as err:
        print(f"Error executing command '{cmd}'\n{err}")
        str_output = ""

    return str_output


def runShellCmdLogging(cmd, shell=True):
    try:
        proc = subprocess.Popen(cmd, shell=shell, stdout=subprocess.PIPE,
                                stderr=subprocess.STDOUT)
        while 1:
            output = proc.stdout.readline().rstrip()
            if output:
                print(output.decode('utf-8', 'backslashreplace'))

            if proc.poll() is not None:
                if proc.returncode != 0:  # must be called *after* poll()
                    print(f"Error executing command '{cmd}'")
                    return 1

                output = proc.stdout.readline().rstrip()
                if output:
                    print(output.decode('utf-8', 'backslashreplace'))
                break

    except (subprocess.CalledProcessError, OSError) as err:
        print(f"Error executing command '{cmd}'\n{err}")
        return 1

    return 0



import pytest
import subprocess

from runner import main as runner
from differ import main as differ
from test_utils import (get_input_path, get_expected_path, get_temp_file_path,
                        generate_ps_dump)

TOOL = 'rotatefont'
CMD = ['-t', TOOL]


# -----
# Tests
# -----

@pytest.mark.parametrize('arg', ['-h', '-v', '-u'])
def test_exit_known_option(arg):
    assert subprocess.call([TOOL, arg]) == 0


@pytest.mark.parametrize('arg', ['-z', '-foo'])
def test_exit_unknown_option(arg):
    assert subprocess.call([TOOL, arg]) == 1


@pytest.mark.parametrize('i, matrix', [
    # scale glyphs horizontally by a factor of 2,
    # condense glyphs vertically by a factor three-quarters,
    # move glyphs left by 100, and up by 200.
    (1, '2.0 0 0 0.75 100 200'),

    # rotate glyphs clockwise 90 degrees around origin point (0,0)
    # move glyphs up by 1000
    (2, '0 -1.0 1.0 0 0 1000'),

    # oblique glyphs, slanting at a slope of 2 units horizontally for every
    # 10 units of height
    (3, '1.0 0 .2 1.0 0 0'),
])
@pytest.mark.parametrize('font_filename', ['font.pfa', 'cidfont.ps'])
def test_matrix(font_filename, i, matrix):
    matrix_values = [f'_{val}' for val in matrix.split()]

    font_path = get_input_path(font_filename)
    actual_path = get_temp_file_path()
    runner(CMD + ['-a', '-f', font_path, actual_path,
                  '-o', 't1', 'matrix'] + matrix_values)
    ext = 'pfa'
    skip = ['-s', '%ADOt1write:']
    if 'cid' in font_filename:
        ext = 'ps'
    expected_path = get_expected_path(f'matrix{i}.{ext}')
    if ext == 'ps':
        actual_path = generate_ps_dump(actual_path)
        expected_path = generate_ps_dump(expected_path)

    assert differ([expected_path, actual_path] + skip)


@pytest.mark.parametrize('i, rt_args', [
    (1, '26 105 -77'),
    (2, '-11 -56 98'),
])
@pytest.mark.parametrize('font_filename', ['font.pfa', 'cidfont.ps'])
def test_rt_fd_options(font_filename, i, rt_args):
    fd_args = []
    ext = 'pfa'
    skip = ['-s', '%ADOt1write:']
    if 'cid' in font_filename:
        fd_args = ['fd', f'_{i}']
        ext = 'ps'

    rt_values = [f'_{val}' for val in rt_args.split()]

    font_path = get_input_path(font_filename)
    actual_path = get_temp_file_path()
    runner(CMD + ['-a', '-f', font_path, actual_path,
                  '-o', 't1', 'rt'] + rt_values + fd_args)
    expected_path = get_expected_path(f'rotatrans{i}.{ext}')
    if ext == 'ps':
        actual_path = generate_ps_dump(actual_path)
        expected_path = generate_ps_dump(expected_path)
    assert differ([expected_path, actual_path] + skip)


@pytest.mark.parametrize('font_filename', ['font.pfa', 'cidfont.ps'])
def test_rtf_option(font_filename):
    ext = 'pfa'
    rtf_filename = 'rotate.txt'
    skip = ['-s', '%ADOt1write:']
    if 'cid' in font_filename:
        ext = 'ps'
        rtf_filename = 'cidrotate.txt'

    rtf_path = get_input_path(rtf_filename)

    font_path = get_input_path(font_filename)
    actual_path = get_temp_file_path()
    runner(CMD + ['-a', '-f', font_path, actual_path,
                  '-o', 't1', 'rt', '_45', '_0', '_0',
                  'rtf', f'_{rtf_path}'])
    expected_path = get_expected_path(f'rotafile.{ext}')
    if ext == 'ps':
        actual_path = generate_ps_dump(actual_path)
        expected_path = generate_ps_dump(expected_path)
    assert differ([expected_path, actual_path] + skip)
